/* GNIO - GLib Network Layer of GIO
 *
 * Copyright (C) 2008 Christian Kellner, Samuel Cormier-Iijima
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Christian Kellner <gicmo@gnome.org>
 *          Samuel Cormier-Iijima <sciyoshi@gmail.com>
 */

#include <config.h>
#include <glib.h>
#include <gio/gio.h>
#include <gio/gasynchelper.h>

#include <string.h>
#ifndef G_OS_WIN32
# include <netinet/in.h>
# include <arpa/inet.h>
# include <netdb.h>
# include <fcntl.h>
# include <unistd.h>
# include <sys/types.h>
#else

#endif
#include <errno.h>

#include "ginetaddress.h"
#include "ginet4address.h"
#include "ginet6address.h"
#include "gsocket.h"
#include "gnioerror.h"
#include "ginetsocketaddress.h"

G_DEFINE_TYPE (GSocket, g_socket, G_TYPE_OBJECT);

enum
{
  PROP_0,
  PROP_FD,
  PROP_BLOCKING
};

struct _GSocketPrivate
{
  gint fd;
  gboolean blocking;
};

static void
g_socket_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
  GSocket *socket = G_SOCKET (object);

  switch (prop_id)
    {
      case PROP_FD:
        g_value_set_int (value, socket->priv->fd);
        break;

      case PROP_BLOCKING:
        g_value_set_boolean (value, socket->priv->blocking);
        break;

      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
g_socket_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
  GSocket *socket = G_SOCKET (object);

  switch (prop_id)
    {
      case PROP_FD:
        socket->priv->fd = g_value_get_int (value);
        break;

      case PROP_BLOCKING:
        g_socket_set_blocking (socket, g_value_get_boolean (value));
        break;

      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
g_socket_finalize (GObject *object)
{
  GSocket *socket G_GNUC_UNUSED = G_SOCKET (object);

  if (G_OBJECT_CLASS (g_socket_parent_class)->finalize)
    (*G_OBJECT_CLASS (g_socket_parent_class)->finalize) (object);
}

static void
g_socket_dispose (GObject *object)
{
  GSocket *socket G_GNUC_UNUSED = G_SOCKET (object);;

  if (G_OBJECT_CLASS (g_socket_parent_class)->dispose)
    (*G_OBJECT_CLASS (g_socket_parent_class)->dispose) (object);
}

static void
g_socket_class_init (GSocketClass *klass)
{
  GObjectClass *gobject_class G_GNUC_UNUSED = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GSocketPrivate));

  gobject_class->finalize = g_socket_finalize;
  gobject_class->dispose = g_socket_dispose;
  gobject_class->set_property = g_socket_set_property;
  gobject_class->get_property = g_socket_get_property;

  g_object_class_install_property (gobject_class, PROP_FD,
                                   g_param_spec_int ("fd",
                                                     "file descriptor",
                                                     "the socket's file descriptor",
                                                     G_MININT,
                                                     G_MAXINT,
                                                     -1,
                                                     G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_BLURB | G_PARAM_STATIC_NICK));

  g_object_class_install_property (gobject_class, PROP_BLOCKING,
                                   g_param_spec_boolean ("blocking",
                                                         "blocking",
                                                         "whether or not this socket is blocking",
                                                         TRUE,
                                                         G_PARAM_READWRITE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_BLURB | G_PARAM_STATIC_NICK));
}

static void
g_socket_init (GSocket *socket)
{
  socket->priv = G_TYPE_INSTANCE_GET_PRIVATE (socket, G_TYPE_SOCKET, GSocketPrivate);

  socket->priv->fd = -1;
  socket->priv->blocking = TRUE;
}

GSocket *
g_socket_new (gint domain, gint type, gint protocol)
{
  gint sock;

  sock = socket(domain, type, protocol);

  if (sock < 0)
    return NULL;

  return G_SOCKET (g_object_new (G_TYPE_SOCKET, "fd", sock, NULL));
}

GSocket *
g_socket_new_from_fd (gint fd)
{
  glong arg;
  gboolean blocking;

  if ((arg = fcntl (fd, F_GETFL, NULL)) < 0)
    g_warning ("Error getting socket status flags: %s", g_strerror (errno));

  blocking = ((arg & O_NONBLOCK) != 0);

  return G_SOCKET (g_object_new (G_TYPE_SOCKET, "blocking", blocking, "fd", fd, NULL));
}

void
g_socket_set_blocking (GSocket  *socket,
                       gboolean  blocking)
{
  glong arg;

  g_return_if_fail (G_IS_SOCKET (socket));

  if ((arg = fcntl (socket->priv->fd, F_GETFL, NULL)) < 0)
    g_warning ("Error getting socket status flags: %s", g_strerror (errno));

  arg = blocking ? arg & ~O_NONBLOCK : arg | O_NONBLOCK;

  if (fcntl (socket->priv->fd, F_SETFL, arg) < 0)
    g_warning ("Error setting socket status flags: %s", g_strerror (errno));

  socket->priv->blocking = blocking;
}

gboolean
g_socket_get_blocking (GSocket *socket)
{
  g_return_val_if_fail (G_IS_SOCKET (socket), FALSE);

  return socket->priv->blocking;
}

GSocketAddress *
g_socket_get_peer_address (GSocket  *socket,
                           GError  **error)
{
  gchar buffer[256];
  gsize len;

  g_return_val_if_fail (G_IS_SOCKET (socket), NULL);

  if (getpeername (socket->priv->fd, (struct sockaddr *) buffer, &len) < 0)
    {
      g_set_error (error, G_IO_ERROR, g_io_error_from_errno (errno), "could not get peer address");
      return NULL;
    }

  return g_socket_address_from_native (buffer, len);
}

void
g_socket_listen (GSocket *socket,
                 gint     backlog)
{
  g_return_if_fail (G_IS_SOCKET (socket));

  listen (socket->priv->fd, backlog);
}

gboolean
g_socket_bind (GSocket         *socket,
               GSocketAddress  *address,
               GError         **error)
{
  g_return_val_if_fail (G_IS_SOCKET (socket) && G_IS_SOCKET_ADDRESS (address), FALSE);

  {
    gchar addr[256];

    if (!g_socket_address_to_native (address, addr))
      return FALSE;

    if (bind (socket->priv->fd, (struct sockaddr *) addr, g_socket_address_native_size (address)) < 0)
      {
        // TODO: set error
        return FALSE;
      }

    g_object_unref (address);

    return TRUE;
  }
}

GSocket *
g_socket_accept (GSocket       *socket,
                 GCancellable  *cancellable,
                 GError       **error)
{
  gint ret;

  if (g_cancellable_set_error_if_cancelled (cancellable, error))
    return NULL;

  if ((ret = accept (socket->priv->fd, NULL, 0)) < 0)
    {
      g_set_error (error, G_IO_ERROR, g_io_error_from_errno (errno), "error accepting connection");
      return NULL;
    }

  if (g_cancellable_set_error_if_cancelled (cancellable, error))
    {
      close (ret);
      return NULL;
    }

  return g_socket_new_from_fd (ret);
}

typedef struct {
  GAsyncReadyCallback callback;
  GCancellable *cancellable;
  gpointer user_data;
  GSocket *socket;
} AcceptData;

static gboolean
accept_callback (AcceptData *data,
                 GIOCondition condition,
                 gint fd)
{
  GSocket *socket;
  GSimpleAsyncResult *result;
  gint ret;

  socket = data->socket;

  if (condition & G_IO_IN)
    {
      if ((ret = accept (socket->priv->fd, NULL, 0)) < 0)
        {
          if (errno == EAGAIN)
            return TRUE;

          result = g_simple_async_result_new_error (G_OBJECT (socket), data->callback, data->user_data, G_IO_ERROR, g_io_error_from_errno (errno), "error accepting connection");
        }
      else
        {
          result = g_simple_async_result_new (G_OBJECT (socket), data->callback, data->user_data, g_socket_accept_async);

          g_simple_async_result_set_op_res_gpointer (result, g_socket_new_from_fd (ret), g_object_unref);
        }
      g_simple_async_result_complete (result);

      g_object_unref (result);
    }

  return FALSE;
}

void
g_socket_accept_async (GSocket             *socket,
                       GCancellable        *cancellable,
                       GAsyncReadyCallback  callback,
                       gpointer             user_data)
{
  GSource *source;
  GSimpleAsyncResult *result;
  AcceptData *data;
  gint ret;

  g_return_if_fail (G_IS_SOCKET (socket));

  if (g_socket_get_blocking (socket))
    g_socket_set_blocking (socket, FALSE);

  if ((ret = accept (socket->priv->fd, NULL, 0)) < 0)
    {
      if (errno == EAGAIN)
        {
          source = _g_fd_source_new (socket->priv->fd, G_IO_IN | G_IO_HUP | G_IO_ERR, cancellable);

          data = g_new (AcceptData, 1);

          data->socket = socket;
          data->callback = callback;
          data->cancellable = cancellable;
          data->user_data = user_data;

          g_source_set_callback (source, (GSourceFunc) accept_callback, data, g_free);

          g_source_attach (source, NULL);
        }
      else
        {
          g_simple_async_report_error_in_idle (G_OBJECT (socket), callback, user_data, G_IO_ERROR, g_io_error_from_errno (errno), "error accepting connection");
        }
    }
  else
    {
      result = g_simple_async_result_new (G_OBJECT (socket), callback, user_data, g_socket_accept_async);

      g_simple_async_result_set_op_res_gpointer (result, g_socket_new_from_fd (ret), g_object_unref);

      g_simple_async_result_complete_in_idle (result);

      g_object_unref (result);
    }
}

GSocket *
g_socket_accept_finish (GSocket       *socket,
                        GAsyncResult  *result,
                        GError       **error)
{
  GSocket *new_socket;
  GSimpleAsyncResult *simple;

  g_return_val_if_fail (G_IS_SOCKET (socket), NULL);

  simple = G_SIMPLE_ASYNC_RESULT (result);

  if (g_simple_async_result_propagate_error (simple, error))
    return NULL;

  g_warn_if_fail (g_simple_async_result_get_source_tag (simple) == g_socket_accept_async);

  new_socket = g_object_ref (g_simple_async_result_get_op_res_gpointer (simple));

  return new_socket;
}

gboolean
g_socket_connect (GSocket         *socket,
                  GSocketAddress  *address,
                  GCancellable    *cancellable,
                  GError         **error)
{
  gchar buffer[256];

  g_return_val_if_fail (G_IS_SOCKET (socket) && G_IS_SOCKET_ADDRESS (address), FALSE);

  g_socket_address_to_native (address, buffer);

  if (connect (socket->priv->fd, (struct sockaddr *) buffer, g_socket_address_native_size (address)) < 0)
    {
      if (errno == EINPROGRESS)
        g_set_error (error, G_IO_ERROR, G_IO_ERROR_PENDING, "connection in progress");
      else
        g_set_error (error, G_IO_ERROR, g_io_error_from_errno (errno), "error connecting: %s", g_strerror (errno));
      return FALSE;
    }

  return TRUE;
}

typedef struct {
  GAsyncReadyCallback callback;
  GCancellable *cancellable;
  gpointer user_data;
  GSocket *socket;
  gchar address_buffer[256];
  gsize address_length;
} ConnectData;

static gboolean
connect_callback (ConnectData *data,
                  GIOCondition condition,
                  gint fd)
{
  GSocket *socket;
  GSimpleAsyncResult *result = NULL;
  gint sockerr = 0;
  gsize sockerr_size = 1;

  socket = data->socket;

  if (condition & G_IO_OUT)
    {
      result = g_simple_async_result_new (G_OBJECT (socket), data->callback, data->user_data, g_socket_connect_async);
    }
  else if (condition & G_IO_ERR)
    {
      if (getsockopt (fd, SOL_SOCKET, SO_ERROR, (gpointer) &sockerr, &sockerr_size) < 0)
        g_warning ("getsockopt: %s", g_strerror (errno));

      if (sockerr != 0)
        result = g_simple_async_result_new_error (G_OBJECT (socket), data->callback, data->user_data, G_IO_ERROR, g_io_error_from_errno (sockerr), "error connecting: %s", g_strerror (sockerr));
      else
        g_warning ("getsockopt SO_ERROR returned no error, with sockerr = %d", sockerr);
    }

  g_simple_async_result_complete (result);

  g_object_unref (result);

  return FALSE;
}

void
g_socket_connect_async (GSocket             *socket,
                        GSocketAddress      *address,
                        GCancellable        *cancellable,
                        GAsyncReadyCallback  callback,
                        gpointer             user_data)
{
  GSource *source;
  GSimpleAsyncResult *result;
  ConnectData *data;
  gint ret;
  gchar buffer[256];
  gsize len;

  g_return_if_fail (G_IS_SOCKET (socket) && G_IS_SOCKET_ADDRESS (address));

  if (g_socket_get_blocking (socket))
    g_socket_set_blocking (socket, FALSE);

  g_socket_address_to_native (address, buffer);

  len = g_socket_address_native_size (address);

  if ((ret = connect (socket->priv->fd, (struct sockaddr *) buffer, len)) < 0)
    {
      if (errno == EINPROGRESS)
        {
          source = _g_fd_source_new (socket->priv->fd, G_IO_OUT | G_IO_ERR, cancellable);

          data = g_new (ConnectData, 1);

          data->socket = socket;
          data->callback = callback;
          data->cancellable = cancellable;
          data->user_data = user_data;
          data->address_length = len;
          memcpy (data->address_buffer, buffer, len);

          g_source_set_callback (source, (GSourceFunc) connect_callback, data, g_free);

          g_source_attach (source, NULL);
        }
      else
        {
          g_simple_async_report_error_in_idle (G_OBJECT (socket), callback, user_data, G_IO_ERROR, g_io_error_from_errno (errno), "error connecting: %s", g_strerror (errno));
        }
    }
  else
    {
      result = g_simple_async_result_new (G_OBJECT (socket), callback, user_data, g_socket_connect_async);

      g_simple_async_result_complete_in_idle (result);

      g_object_unref (result);
    }
}

gboolean
g_socket_connect_finish (GSocket       *socket,
                         GAsyncResult  *result,
                         GError       **error)
{
  GSimpleAsyncResult *simple;

  g_return_val_if_fail (G_IS_SOCKET (socket), FALSE);

  simple = G_SIMPLE_ASYNC_RESULT (result);

  if (g_simple_async_result_propagate_error (simple, error))
    return FALSE;

  g_warn_if_fail (g_simple_async_result_get_source_tag (simple) == g_socket_connect_async);

  return TRUE;
}
